/****************************************************************************
** $Id: qt/gltexobj.cpp   3.1.2   edited Nov 8 2002 $
**
** Copyright (C) 1992-2002 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

/****************************************************************************
**
** This is a simple QGLWidget demonstrating the use of QImages for textures.
**
** Much of the GL code is inspired by the 'spectex' and 'texcyl' 
** public domain demo programs by Brian Paul.
**
****************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <list>
#include <algorithm>
#include <cmath>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#include <QImage>
#include <QTimer>
#include <QEvent>
#include <QInputEvent>
#include <QString>
#include <QFileDialog>
#include <QMouseEvent>
#include <QMessageBox>

#include "point.h"
#include "kdtree.h"
#include "gltexobj.h"

const int redrawWait = 50;
//const int redrawWait = 0;

/*!
  Create a GLTexobj widget
*/

GLTexobj::GLTexobj( QWidget* parent ) :
	QGLWidget( parent ),
	img(width(),height(),QImage::Format_RGB32),
    	textfont(),
	x(0.0),y(0.0),
	drawingBox(false),
	query(NULL),
	nearest(NULL),
	radius(0.0),
	k(3)
{
  timer = new QTimer( this );
  connect( timer, SIGNAL(timeout()), SLOT(update()) );
  timer->setInterval( redrawWait );
  timer->setSingleShot( FALSE );
  timer->start();

  // looks like font selection under X11 may be limited...
  textfont.setFamily("Fixed");
  textfont.setRawMode(true);
  textfont.setStyleStrategy(QFont::OpenGLCompatible);
//textfont.setPixelSize(10);
  textfont.setPointSize(10);
  textfont.setFixedPitch(true);
  textfont.setStyleHint(QFont::AnyStyle,QFont::PreferBitmap);

  QFontInfo	textfontinfo(textfont);

  std::cout << "family: " << textfontinfo.family().toStdString() << " ";
  std::cout << "pointsize: " << textfontinfo.pointSize() << " ";
  std::cout << "weight: " << textfontinfo.weight() << " ";
  std::cout << "italic: " << textfontinfo.italic() << " ";
  std::cout << std::endl;

  clearAll();
}


/*!
  Release allocated resources
*/

GLTexobj::~GLTexobj()
{
  makeCurrent();
}


/*!
  Set up the OpenGL rendering state, and define display list
*/

void GLTexobj::initializeGL()
{
  // specify clear color; (0.3,0.4,0.6,1.0) is a nice "slate blue"
  glClearColor( 0.0, 0.0, 0.0, 0.0 );

  std::cerr << "GL: " << glGetString(GL_VERSION) << std::endl;
}


/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void GLTexobj::resizeGL( int w, int h )
{
  glViewport(0,0,w,h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
//gluOrtho2D(0.0,w,0.0,h);
  gluOrtho2D(-w,w,-h,h);
//gluPerspective(90,1.0,1.0,1000.0);
//glFrustum(-1.0,1.0,-1.0,1.0,1.0,1000.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


/*!
  Paint the texobj. The actual openGL commands for drawing the texobj are
  performed here.
*/

void GLTexobj::paintGL()
{
	ostringstream	os;
	GLUquadricObj*	qobj=NULL;

  if(!qobj) qobj = gluNewQuadric();

//std::cerr << "(" << x << "," << y << ")" << std::endl;

  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // render here
  glColor4f(1.0,1.0,1.0,1.0);
  os << "(" << x << "," << y << ")";
  renderText(x-15,y+15,0.0,QString(os.str().c_str()),textfont);

  // draw white box at (moving) mouse coordinates
  glColor4f(1.0,1.0,1.0,1.0);
  glBegin(GL_QUADS);
  glVertex2f(x-5,y-5);
  glVertex2f(x+5,y-5);
  glVertex2f(x+5,y+5);
  glVertex2f(x-5,y+5);
  glEnd();

  // draw points (in green)
  glColor4f(0.0, 1.0, 0.0, 1.0);
  for(int i=0; i < (int)pts.size(); ++i) {
    glBegin(GL_QUADS);
    glVertex2f((*pts[i])[0]-5.0,(*pts[i])[1]-5.0);
    glVertex2f((*pts[i])[0]+5.0,(*pts[i])[1]-5.0);
    glVertex2f((*pts[i])[0]+5.0,(*pts[i])[1]+5.0);
    glVertex2f((*pts[i])[0]-5.0,(*pts[i])[1]+5.0);
    glEnd();
  }

  kdtree.render();

  // draw nn query point
  glColor4f(1.0, 0.0, 0.0, 1.0);
  if(query) {
    // draw query point
    glBegin(GL_QUADS);
    glVertex2f((*query)[0]-5.0,(*query)[1]-5.0);
    glVertex2f((*query)[0]+5.0,(*query)[1]-5.0);
    glVertex2f((*query)[0]+5.0,(*query)[1]+5.0);
    glVertex2f((*query)[0]-5.0,(*query)[1]+5.0);
    glEnd();

    // draw query radius
  //std::cerr << "radius = " << radius << std::endl;
    glColor4f(1.0, 1.0, 0.0, 1.0);
    glPushMatrix();
    glTranslatef((*query)[0],(*query)[1],0.0);
    gluDisk(qobj,radius-2.0,radius,360,1);
    glPopMatrix();
  }
  // draw query result: either nn point or k-nn points
  glColor4f(0.0, 0.0, 1.0, 1.0);
  if(nearest) {
    // draw nn point
    glBegin(GL_QUADS);
    glVertex2f((*nearest)[0]-5.0,(*nearest)[1]-5.0);
    glVertex2f((*nearest)[0]+5.0,(*nearest)[1]-5.0);
    glVertex2f((*nearest)[0]+5.0,(*nearest)[1]+5.0);
    glVertex2f((*nearest)[0]-5.0,(*nearest)[1]+5.0);
    glEnd();
  }
  // draw knn points
  for(int i=0; i < (int)knearest.size(); ++i) {
    glBegin(GL_QUADS);
    glVertex2f((*knearest[i])[0]-5.0,(*knearest[i])[1]-5.0);
    glVertex2f((*knearest[i])[0]+5.0,(*knearest[i])[1]-5.0);
    glVertex2f((*knearest[i])[0]+5.0,(*knearest[i])[1]+5.0);
    glVertex2f((*knearest[i])[0]-5.0,(*knearest[i])[1]+5.0);
    glEnd();
  }

  // draw the range query bounding box
  glColor4f(1.0, 1.0, 0.0, 1.0);
  glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  if(drawingBox) glRectd(x1,y1,x2,y2);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

  // draw the resulting points
  glColor4f(0.0, 0.0, 1.0, 1.0);
  for(int i=0; i < (int)range.size(); ++i) {
    glBegin(GL_QUADS);
    glVertex2f((*range[i])[0]-5.0,(*range[i])[1]-5.0);
    glVertex2f((*range[i])[0]+5.0,(*range[i])[1]-5.0);
    glVertex2f((*range[i])[0]+5.0,(*range[i])[1]+5.0);
    glVertex2f((*range[i])[0]-5.0,(*range[i])[1]+5.0);
    glEnd();
  }

  // swapbuffers
//swapBuffers();
}

void GLTexobj::clearQuery()
{
  if(query) { delete query;  query = NULL; }
  knearest.clear();
  nearest = NULL;
}

void GLTexobj::clearAll()
{
	std::vector<point_t*>::iterator	pit;

  if(query) { delete query;  query = NULL; }
  knearest.clear();
  nearest = NULL;

  kdtree.clear();
  for(pit=pts.begin(); pit!=pts.end(); pit=pts.erase(pit))
    if(*pit) delete *pit;
}

void GLTexobj::growkdtree()
{
  // create kd-tree
  kdtree.insert(pts,point_t(-width(),-height()),point_t(width(),height()));

//std::cerr << "kd-tree (in-order): " << std::endl << kdtree;
}

void GLTexobj::setknnval(const QString& str)
{
  k = str.toInt() > 0 ? str.toInt() : 1;
}

void GLTexobj::mousePressEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;

  mapCoords(e->x(),e->y());

  if(e->button() == Qt::LeftButton) {
    if(e->modifiers() == Qt::NoModifier) {
      pts.push_back(new point_t(x,y));
    } else
    if(e->modifiers() == Qt::AltModifier) {
      //  nearest neighbor
    } else
    if(e->modifiers() == (Qt::AltModifier | Qt::ShiftModifier)) {
      //  k nearest neighbor
    } else
    if(e->modifiers() == Qt::ShiftModifier) {
      // starting to draw the range query bounding box
      drawingBox = true;
      x1 = x; y1 = y;
      x2 = x; y2 = y;
    }
  }

  e->accept();
  updateGL();
}

void GLTexobj::mouseReleaseEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;

  mapCoords(e->x(),e->y());

  // reset previous queries, if any
  if(query) { delete query;  query = NULL; }
  nearest = NULL;
  knearest.clear();
  range.clear();

  if(e->button() == Qt::LeftButton) {
    if(e->modifiers() == Qt::NoModifier) {
      // point insertion (no-op here)
    } else
    if(e->modifiers() == Qt::AltModifier) {
      //  nearest neighbor
      query = new point_t(x,y);
      kdtree.nn(query,nearest,radius);
    } else
    if(e->modifiers() == (Qt::AltModifier | Qt::ShiftModifier)) {
      //  k nearest neighbor
      query = new point_t(x,y);
      kdtree.knn(query,knearest,radius,k);
    } else
    if(e->modifiers() == Qt::ShiftModifier) {
      // finished drawing the range query bounding box
      x2 = x; y2 = y;
      point_t min,max;
      min[0] = x1 < x2 ? x1 : x2;  max[0] = x1 > x2 ? x1 : x2;
      min[1] = y1 < y2 ? y1 : y2;  max[1] = y1 > y2 ? y1 : y2;

      drawingBox = false;
      kdtree.range(min,max,range);
//    std::cerr << "Range query returned " << range.size() << " points\n";
    }
  }

  e->accept();
  updateGL();
}

void GLTexobj::mouseDoubleClickEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;
//std::cerr << "mouseDoubleClick" << std::endl;

  e->accept();
  updateGL();
}

void GLTexobj::wheelEvent(QWheelEvent* e)
{
        float   degrees = ((float)e->delta() / 360.0); // should be WHEEL_DELTA
//      float   radians = degrees * M_PI / 180.0;

  std::cerr << "wheelEvent:";
  std::cerr << " delta: " << e->delta();
  std::cerr << " degrees: " << degrees << std::endl;
  e->accept();
  updateGL();
}

void GLTexobj::mouseMoveEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;

  mapCoords(e->x(),e->y());

  // update current bounding box second coordinates
  x2 = x;
  y2 = y;

  e->accept();
  updateGL();
}

void GLTexobj::mapCoords(int ex, int ey)
{
  x = ex;
  y = height() - ey;

  double dx = (double)x/width();
  double dy = (double)y/height();

  dx = width()*(2.0*dx - 1.0);
  dy = height()*(2.0*dy - 1.0);

  x = dx;
  y = dy;
}

void GLTexobj::update()
{
  updateGL();
}


void GLTexobj::fileOpen()
{
	point_t*	ptp;
	QString qfilename =
		QFileDialog::getOpenFileName(this,tr("Open File"),
                                             "./",QString::null);

  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.toStdString().c_str());
    if(!ifs)
      std::cerr << "Warning: can't open " << qfilename.toStdString() << std::endl;
    else {
      // shorthanded version
      while(ifs >> (ptp = new point_t(0.0,0.0))) pts.push_back(ptp);
      // this works too
//    while(!ifs.eof()) {
//      ptp = new point_t(0.0,0.0);
//      ifs >> ptp;
//      point.push_back(ptp);
//    }
    }
    ifs.close();
  }
}


void GLTexobj::fileSaveAs()
{
	std::vector<point_t*>::iterator	pit;
	QString qfilename =
		QFileDialog::getSaveFileName(this,tr("Save File"),
                                             "./",tr("Points (*.pts)"));

  if(!qfilename.isEmpty()) {
    std::ofstream ofs(qfilename.toStdString().c_str());
    for(pit = pts.begin(); pit != pts.end(); ++pit) ofs << *pit;
    ofs.close();
  }
}


void GLTexobj::imgSaveAs()
{
  glReadBuffer(GL_BACK);
  img = QGLWidget::grabFrameBuffer(true);
	QString qfilename =
		QFileDialog::getSaveFileName(this,tr("Save File"),
                                             "./",tr("Images (*.png)"));

  img.save(qfilename,"PNG");
}

void GLTexobj::about()
{
  QMessageBox::about(this, tr("KdTree"),
  tr(
    "<small>"
    "Usage:"
    "<p>"
    // 3-button mouse
  //"<table>"
  //"<tr><td>left mouse</td><td>point insertion</td></tr>"
  //"<tr><td>middle mouse</td><td>nearest-neighbor query</td></tr>"
  //"<tr><td>CTRL middle mouse</td><td>k-nearest-neighbor query</td></tr>"
  //"<tr><td>right mouse</td><td>range query</td></tr>"
  //"</table>"
  //"<p>"
    // trackpad
    "<table>"
    "<tr>"
      "<td align=right><tt>left mouse:</tt></td>"
      "<td>&nbsp;</td>"
      "<td><i>point insertion</i></td>"
    "</tr>"
    "<tr>"
      "<td align=right><tt>ALT left mouse:</tt></td>"
      "<td>&nbsp;</td>"
      "<td><i>nearest-neighbor query</i></td>"
    "</tr>"
    "<tr>"
      "<td align=right><tt>ALT SHIFT left mouse:</tt></td>"
      "<td>&nbsp;</td>"
      "<td><i>k-nearest-neighbor query</i></td>"
    "</tr>"
    "<tr>"
      "<td align=right><tt>SHIFT left mouse:</tt></td>"
      "<td>&nbsp;</td>"
      "<td><i>range query</i></td>"
    "</tr>"
    "</table>"
    "<p>"
  ));
}
